/* The MazezaM Engine 
 * Copyright (C) 2002-2008 Malcolm Tyrrell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package file;

import globals.Globals;

import java.util.Iterator;
import java.util.LinkedList;

import level.FloorGrid;
import level.Level;
import level.Unit;
import level.UnitShapeGroup;
import level.UnitShapePosPair;
import level.Wall;
import tessellation.NeighbourIterator;
import tessellation.Position;
import tessellation.TessellationManager;

/**
 * LevelParsers are provided to other parsers to allow them to parse the level
 * portion of their files.
 */
public abstract class LevelParser {
	/**
	 * We should be able to use this to construct an XML tag.
	 * 
	 * e.g. "Sokoban" in "SokobanLevels"
	 */
	protected abstract String getGameTitle();

	/**
	 * Returns an extension suitable for standard level parsing
	 * 
	 * e.g. "mzm" for "MazezaM"
	 */
	protected abstract String getExtension();

	LevelFileFilter get_filter() {
		return new LevelFileFilter(getExtension(), getExtension().toUpperCase() 
				+ Globals.text.getString("file_filter_files") + " (" + getGameTitle() + ")");
	}

	/**
	 * Does the string represent a line of a level
	 */
	protected abstract boolean isLevelLine(String s);

	/**
	 * Parses a string array and returns a single level; levelNo is provided to
	 * allow for graphical variety
	 */
	protected abstract Level parseLevel(String t, String a, String[] sa,
			int levelNo, int levelDepth) throws LevelParseException;

	/**
	 * This is handy if you have a list.
	 */
	protected Level parseLevel(String t, String a, LinkedList<String> sl,
			int levelNo, int levelDepth) throws LevelParseException {
		String[] levelStrings = new String[sl.size()];
		Iterator<String> sli;
		int i = 0;
		for (sli = sl.iterator(); sli.hasNext(); i++) {
			levelStrings[i] = sli.next();
		}
		return parseLevel(t, a, levelStrings, levelNo, levelDepth);
	}

	/*
	 * LOTS OF USEFUL METHODS FOR SUBCLASSES TO USE
	 */

	/**
	 * returns a UnitShapeGroup containing the walls
	 */
	protected static UnitShapeGroup parseWallsFromStringArray(String[] sa) {
		UnitShapeGroup usg = new UnitShapeGroup();
		int i, j;
		// loop over array looking for wall units. If wall unit doesn't already
		// belong to an identified wall object, add the new wall object
		// which should contain it.
		for (j = 0; j < sa.length; j++)
			for (i = 0; i < sa[j].length(); i++)
				if ((sa[j].charAt(i) == Unit.wall)
						&& (usg.empty(TessellationManager.newPosition(i, j)))) {
					usg.add(new UnitShapePosPair(wallAtPos(sa, i, j),
							TessellationManager.newPosition()));
				}
		return usg;
	}

	/**
	 * returns the Wall object which contains a unit at the given coordinates
	 */
	private static Wall wallAtPos(String[] sa, int x, int y) {
		// I should make sure that the strings in sa are non-empty)
		// String[] new_sa = new String[sa.length];
		// int i, j;
		LinkedList<Position> the_wall = new LinkedList<Position>();
		wallAtPos_rec(sa, TessellationManager.newPosition(x, y), the_wall);
		return new Wall(the_wall);
	}

	/**
	 * Is there a wall unit at p
	 */
	private static boolean wall_at(String[] sa, Position p) {
		int x = p.array_x();
		int y = p.array_y();
		if ((y >= 0) && (y < sa.length))
			if ((x >= 0) && (x < sa[y].length()))
				return sa[y].charAt(x) == Unit.wall;
		return false;
	}

	/**
	 * recursive algorithm which builds a wall
	 */
	private static void wallAtPos_rec(String[] sa, Position p,
			LinkedList<Position> pl) {
		NeighbourIterator ni;
		Position p_neighbour;
		pl.add(p);
		for (ni = new NeighbourIterator(p); ni.hasNext();) {
			p_neighbour = ni.next();
			if ((wall_at(sa, p_neighbour)) && (!(pl.contains(p_neighbour))))
				wallAtPos_rec(sa, p_neighbour, pl);
		}
	}

	/**
	 * Parse the floor from a string array
	 */
	protected static FloorGrid parseBlankFloorFromStringArray(String[] sa,
			Position pos) {
		LinkedList<Position> pl = new LinkedList<Position>();
		blankTilesAtPos_rec(sa, pos, pl);
		return new FloorGrid(pl);
	}

	/*
	 * Helper function - we can't simply negate wall_at, since both methods have
	 * to reject units outside the array.
	 */
	private static boolean non_wall_at(String[] sa, Position p) {
		int x = p.array_x();
		int y = p.array_y();
		if ((y >= 0) && (y < sa.length))
			if ((x >= 0) && (x < sa[y].length()))
				return sa[y].charAt(x) != Unit.wall;
		return false;
	}

	protected static void blankTilesFromPos(String[] sa, Position pos, LinkedList<Position> pl) {
		if (! pl.contains(pos)) 
			blankTilesAtPos_rec(sa,pos,pl);
	}
	
	private static void blankTilesAtPos_rec(String[] sa, Position p,
			LinkedList<Position> pl) {
		NeighbourIterator ni;
		Position p_neighbour;
		pl.add(p);
		for (ni = new NeighbourIterator(p); ni.hasNext();) {
			p_neighbour = ni.next();
			if ((non_wall_at(sa, p_neighbour)) && (!(pl.contains(p_neighbour))))
				blankTilesAtPos_rec(sa, p_neighbour, pl);
		}
	}
}
